import jax
import jax.numpy as jnp
from jax import grad, jit
from jax.lax import map
from jaxfin.price_engine.math.norm import cum_normal


@jit
def compute_d1(spots_1, spots_2, sigma_sqrt_t):
    return (jnp.log(spots_2 / spots_1) / sigma_sqrt_t) + 0.5 * sigma_sqrt_t


@jit
def compute_d1_d2(spots_1, spots_2, expires, sigma_1, sigma_2, corr):
    sigma = jnp.sqrt(
        jnp.sum(jnp.asarray([sigma_1, sigma_2]) ** 2)
        - 2 * corr * jnp.prod(jnp.asarray([sigma_1, sigma_2]))
    )

    sigma_sqrt_t = sigma * jnp.sqrt(expires)
    d1 = compute_d1(spots_1, spots_2, sigma_sqrt_t)
    d2 = d1 - sigma_sqrt_t

    return d1, d2


@jit
def margrabe(
    spots_1: jax.Array,
    spots_2: jax.Array,
    expires: jax.Array,
    sigma_1: jax.Array,
    sigma_2: jax.Array,
    corr: jax.Array,
):
    """
    Calculate the price of a spread option max(S_2 - S_1, 0) using the Margrabe formula.
    Both S_2 and S_1 are assumed to be log-normally distributed with correlation corr.

    :param spots_1: The spot price of the first asset
    :param spots_2: The spot price of the second asset
    :param expires: The time to expiration of the option
    :param sigma_1: The volatility of the first asset
    :param sigma_2: The volatility of the second asset
    :param corr: The correlation between the two assets
    :param dtype: The dtype of the input
    :return: The price of the spread option
    """
    d1, d2 = compute_d1_d2(spots_1, spots_2, expires, sigma_1, sigma_2, corr)

    return spots_2 * cum_normal(d1) - spots_1 * cum_normal(d2)


@jit
def margrabe_deltas(spots_1, spots_2, expires, sigma_1, sigma_2, corr) -> jax.Array:
    """
    Calculate the deltas (with respect to S1 and S2 respectively) of a spread option max(S_2 - S_1, 0) using the Margrabe formula.

    :param spots_1: The spot price of the first asset
    :param spots_2: The spot price of the second asset
    :param expires: The time to expiration of the option
    :param vols: The volatility of the two assets
    :param corr: The correlation between the two assets
    :return: The deltas of the spread option
    """
    delta_1 = grad(margrabe, argnums=0)(
        spots_1, spots_2, expires, sigma_1, sigma_2, corr
    )
    delta_2 = grad(margrabe, argnums=1)(
        spots_1, spots_2, expires, sigma_1, sigma_2, corr
    )

    return jnp.asarray([delta_1, delta_2])


@jit
def margrabe_gammas(spots_1, spots_2, expires, sigma_1, sigma_2, corr):
    """

    :param spots_1:
    :param spots_2:
    :param expires:
    :param vols:
    :param corr:
    :param exchange_first:
    :param dtype:
    :return:
    """
    gamma_1 = grad(grad(margrabe, argnums=0), argnums=0)(
        spots_1, spots_2, expires, sigma_1, sigma_2, corr
    )
    gamma_2 = grad(grad(margrabe, argnums=1), argnums=1)(
        spots_1, spots_2, expires, sigma_1, sigma_2, corr
    )

    return jnp.asarray([gamma_1, gamma_2])


@jit
def margrabe_cross_gamma(spots_1, spots_2, expires, sigma_1, sigma_2, corr):
    """

    :param spots_1:
    :param spots_2:
    :param expires:
    :param vols:
    :param corr:
    :param exchange_first:
    :param dtype:
    :return:
    """
    cross_gamma_1 = grad(grad(margrabe, argnums=0), argnums=1)(
        spots_1, spots_2, expires, sigma_1, sigma_2, corr
    )
    cross_gamma_2 = grad(grad(margrabe, argnums=1), argnums=0)(
        spots_1, spots_2, expires, sigma_1, sigma_2, corr
    )

    return jnp.asarray([cross_gamma_1, cross_gamma_2])


def v_margrabe(
    spots_1: jax.Array,
    spots_2: jax.Array,
    expires: jax.Array,
    sigma_1: jax.Array,
    sigma_2: jax.Array,
    corr: jax.Array,
):
    """
    Vectorized version of the Margrabe formula.

    :param spots_1:
    :param spots_2:
    :param expires:
    :param sigma_1:
    :param sigma_2:
    :param corr:
    :param exchange_first:
    :param dtype:
    :return:
    """
    if spots_1.shape == ():
        spots_1 = jnp.expand_dims(spots_1, axis=0)
        spots_2 = jnp.expand_dims(spots_2, axis=0)
        expires = jnp.expand_dims(expires, axis=0)
        sigma_1 = jnp.expand_dims(sigma_1, axis=0)
        sigma_2 = jnp.expand_dims(sigma_2, axis=0)
        corr = jnp.expand_dims(corr, axis=0)

    return map(
        lambda spots: margrabe(spots[0], spots[1], expires, sigma_1, sigma_2, corr),
        (spots_1, spots_2),
    )



import gymnasium
import numpy as np
from jaxfin.price_engine.black_scholes import delta_european, european_price

from hedging_gym.utils.env_checker import check_env

s0 = 100.0
strike = 100.0
expiry = 1.0
r = 0.0
mu = 0.0
sigma = 0.2
n_steps = 252
dt = expiry / n_steps

SEED = 0


def test_check_env_cont():
    env = gymnasium.make(
        "CallHedgingBSCont-v0",
        s0=s0,
        strike=strike,
        expiry=expiry,
        r=r,
        mu=mu,
        sigma=sigma,
        n_steps=n_steps,
    )

    try:
        check_env(env)
    except Exception as e:
        assert False, f"check_env raised an exception: {e}"

    assert True


def test_check_env_dis():
    env = gymnasium.make(
        "CallHedgingBSDiscrete-v0",
        s0=s0,
        strike=strike,
        expiry=expiry,
        r=r,
        mu=mu,
        sigma=sigma,
        n_steps=n_steps,
    )

    try:
        check_env(env)
    except Exception as e:
        assert False, f"check_env raised an exception: {e}"

    assert True


def test_reset_cont():
    env = gymnasium.make(
        "CallHedgingBSCont-v0",
        s0=s0,
        strike=strike,
        expiry=expiry,
        r=r,
        mu=mu,
        sigma=sigma,
        n_steps=n_steps,
    )
    obs, info = env.reset(seed=SEED)

    bs_delta_0 = delta_european(s0, strike, expiry, sigma, r)
    call_price_0 = european_price(s0, strike, expiry, sigma, r)

    expected_info = {
        "price": call_price_0,
        "time_to_expiration": 1.0,
        "bs_delta": bs_delta_0,
        "stock_price": s0,
        "current_delta": -bs_delta_0,
        "log(S/K)": np.log(s0 / strike),
        "hedge_portfolio_value": call_price_0,
        "bank_account": 46.02264887044997,
    }
    expected_obs = np.array(
        [0.0, sigma, 1.0, bs_delta_0, 1.0, -bs_delta_0], dtype=np.float32
    )

    assert np.array_equal(obs, expected_obs)
    for key in expected_info:
        assert np.isclose(info[key], expected_info[key])


def test_reset_dis():
    env = gymnasium.make(
        "CallHedgingBSDiscrete-v0",
        s0=s0,
        strike=strike,
        expiry=expiry,
        r=r,
        mu=mu,
        sigma=sigma,
        n_steps=n_steps,
    )
    obs, info = env.reset(seed=SEED)

    bs_delta_0 = 0.5398278
    call_price_0 = 7.965561

    expected_info = {
        "price": call_price_0,
        "time_to_expiration": 1.0,
        "bs_delta": bs_delta_0,
        "stock_price": s0,
        "current_delta": -bs_delta_0,
        "log(S/K)": np.log(s0 / strike),
        "hedge_portfolio_value": call_price_0,
        "bank_account": 46.02264887044997,
    }
    expected_obs = np.array(
        [0.0, sigma, 1.0, bs_delta_0, 1.0, -bs_delta_0], dtype=np.float32
    )

    assert np.array_equal(obs, expected_obs)
    for key in expected_info:
        assert np.isclose(info[key], expected_info[key])


def test_step():
    env = gymnasium.make(
        "CallHedgingBSCont-v0",
        s0=s0,
        strike=strike,
        expiry=expiry,
        r=r,
        mu=mu,
        sigma=sigma,
        n_steps=n_steps,
    )
    obs, info = env.reset(seed=SEED)
    dt = expiry / n_steps
    stock_path = env.unwrapped.stock_path[:, 0]

    expected_call_prices = np.asarray(
        [
            european_price(s, strike, expiry - i * dt, sigma, r)
            for i, s in enumerate(stock_path)
        ]
    )
    expected_deltas = np.asarray(
        [
            delta_european(s, strike, expiry - i * dt, sigma, r)
            for i, s in enumerate(stock_path)
        ]
    )

    call_prices = [info["price"]]
    bs_deltas = [info["bs_delta"]]

    for i in range(252):
        action = env.action_space.sample()

        obs, rewards, done, _, info = env.step(action)

        call_prices.append(info["price"])
        bs_deltas.append(info["bs_delta"])

        if done:
            break

    np.allclose(np.asarray(call_prices), expected_call_prices)
    np.allclose(np.asarray(bs_deltas), expected_deltas)



import gymnasium
import numpy as np

from hedging_gym.envs.margrabe_env import MargrabeEnvCont
from hedging_gym.math import margrabe, margrabe_deltas
from hedging_gym.utils.env_checker import check_env

s1_0 = 100.0
s2_0 = 100.0
expiry = 1.0
r = 0.0
mu_1 = 0.0
mu_2 = 0.0
sigma_1 = 0.2
sigma_2 = 0.2
corr = 0.5
n_steps = 252
dt = expiry / n_steps

SEED = 0

np.random.seed(SEED)


def test_check_env_cont():
    env = gymnasium.make(
        "MargrabeHedgingCont-v0",
        s1_0=s1_0,
        s2_0=s2_0,
        expiry=expiry,
        r=r,
        mu_1=mu_1,
        mu_2=mu_2,
        sigma_1=sigma_1,
        sigma_2=sigma_2,
        corr=corr,
        n_steps=n_steps,
    )

    try:
        check_env(env)
    except Exception as e:
        assert False, f"check_env raised an exception: {e}"

    assert True


def test_reset_cont():
    env = gymnasium.make(
        "MargrabeHedgingCont-v0",
        s1_0=s1_0,
        s2_0=s2_0,
        expiry=expiry,
        r=r,
        mu_1=mu_1,
        mu_2=mu_2,
        sigma_1=sigma_1,
        sigma_2=sigma_2,
        corr=corr,
        n_steps=n_steps,
    )
    obs, info = env.reset(seed=SEED)

    deltas = margrabe_deltas(s1_0, s2_0, expiry, sigma_1, sigma_2, corr)
    m_delta_1, m_delta_2 = deltas[0], deltas[1]
    spread_price = margrabe(s1_0, s2_0, expiry, sigma_1, sigma_2, corr)
    log_ratio = np.log(s2_0 / s1_0)

    expected_info = {
        "price": spread_price,
        "time_to_expiration": 1.0,
        "m_delta_1": m_delta_1,
        "m_delta_2": m_delta_2,
        "stock_1": s1_0,
        "stock_2": s2_0,
        "current_delta_1": m_delta_1,
        "current_delta_2": m_delta_2,
        "log(S_2/S_1)": log_ratio,
        "hedge_portfolio_value": spread_price,
        "bank_account": -15.921075323618034,
    }
    expected_obs = np.array(
        [
            log_ratio,
            sigma_1,
            sigma_2,
            1.0,
            m_delta_1,
            m_delta_2,
            1.0,
            m_delta_1,
            m_delta_2,
        ],
        dtype=np.float32,
    )

    assert np.array_equal(obs, expected_obs)
    for key in expected_info:
        assert np.isclose(info[key], expected_info[key])


def test_margrabe_step():
    env = gymnasium.make(
        "MargrabeHedgingCont-v0",
        s1_0=s1_0,
        s2_0=s2_0,
        expiry=expiry,
        r=r,
        mu_1=mu_1,
        mu_2=mu_2,
        sigma_1=sigma_1,
        sigma_2=sigma_2,
        corr=corr,
        n_steps=n_steps,
    )
    obs, info = env.reset(seed=SEED)
    dt = expiry / n_steps
    stock_path_1, stock_path_2 = env.unwrapped.stock_path_1, env.unwrapped.stock_path_2

    expected_spread_prices = np.asarray(
        [
            margrabe(
                stock_path_2[i],
                stock_path_1[i],
                expiry - i * dt,
                sigma_1,
                sigma_2,
                corr,
            )
            for i in range(n_steps)
        ]
    ).flatten()
    expected_deltas = np.asarray(
        [
            margrabe_deltas(
                stock_path_2[i, 0],
                stock_path_1[i, 0],
                expiry - i * dt,
                sigma_1,
                sigma_2,
                corr,
            )
            for i in range(n_steps)
        ]
    )
    expected_delta_1 = expected_deltas[:, 0]
    expected_delta_2 = expected_deltas[:, 1]

    spread_prices = [info["price"]]
    deltas_1 = [info["m_delta_1"]]
    deltas_2 = [info["m_delta_2"]]

    for _ in range(252):
        action = np.array([0.0, 0.0], dtype=np.float32)
        obs, reward, done, _, info = env.step(action)

        spread_prices.append(info["price"])
        deltas_1.append(info["m_delta_1"])
        deltas_2.append(info["m_delta_2"])

        if done:
            break

    assert np.allclose(np.asarray(spread_prices), expected_spread_prices)
    assert np.allclose(np.asarray(deltas_1), expected_delta_1)
    assert np.allclose(np.asarray(deltas_2), expected_delta_2)



from typing import Union

import numpy as np
from gymnasium import spaces
from gymnasium.spaces import Box, Discrete
from jax import vmap
from jaxfin.models.heston import UnivHestonModel
from jaxfin.price_engine.fft import delta_call_fourier, fourier_inv_call

from ..base.call_env import HedgingEnvBase

v_delta_call_fourier = vmap(
    delta_call_fourier, in_axes=(0, None, None, None, None, None, None, None, None)
)
v_fouirer_inv_call = vmap(
    fourier_inv_call, in_axes=(0, None, None, None, None, None, None, None, None)
)


def flatten(fn):
    def wrapper(*args, **kwargs):
        return fn(*args, **kwargs).flatten()

    return wrapper


class HestonEnvBase(HedgingEnvBase):
    metadata = {"render_modes": ["human"]}
    action_space: spaces.Space
    observation_space: spaces.Space

    def __init__(
        self,
        s0: float,
        strike: float,
        expiry: float,
        r: float,
        mu: float,
        v0: float,
        kappa: float,
        theta: float,
        sigma: float,
        rho: float,
        n_steps: int,
        ticksize: float = 0.01,
    ):
        super().__init__(
            s0=s0,
            strike=strike,
            expiry=expiry,
            r=r,
            mu=mu,
            sigma=theta,
            n_steps=n_steps,
            ticksize=ticksize,
        )
        self.v0 = v0
        self.kappa = kappa
        self.theta = theta
        self.rho = rho
        self.vol_of_vol = sigma
        self._variance_process = np.asarray([])

    def _generate_stock_path(self) -> np.ndarray:
        heston_process = UnivHestonModel(
            s0=self.s0,
            v0=self.v0,
            mean=self.mu,
            kappa=self.kappa,
            theta=self.theta,
            sigma=self.vol_of_vol,
            rho=self.rho,
        )

        paths, variance_p = heston_process.sample_paths(self.expiry, self.n_steps, 1)
        paths_np = np.asarray(paths)
        variance_p = np.asarray(variance_p).squeeze()  # type: ignore
        self._variance_process = variance_p  # type: ignore

        return paths_np

    def _get_current_stock_vol(self, step: int) -> float:
        return np.sqrt(self._variance_process[step])

    @flatten
    def _get_call_prices(self) -> np.ndarray:
        return np.asarray(
            [
                fourier_inv_call(
                    s0=self._stock_path[i],
                    K=self.strike,
                    T=self.expiry - i * self.dt,
                    v0=self._variance_process[i],
                    mu=self.mu,
                    kappa=self.kappa,
                    theta=self.theta,
                    sigma=self.vol_of_vol,
                    rho=self.rho,
                )
                for i in range(self.n_steps)
            ]
        )

    @flatten
    def _get_deltas(self) -> np.ndarray:
        return np.asarray(
            [
                v_delta_call_fourier(
                    self._stock_path[i],
                    self.strike,
                    self.expiry - i * self.dt,
                    self._variance_process[i],
                    self.mu,
                    self.theta,
                    self.vol_of_vol,
                    self.kappa,
                    self.rho,
                )
                for i in range(self.n_steps)
            ]
        )
    
    @property
    def variance_process(self):
        return self._variance_process


class HestonEnvCont(HestonEnvBase):

    def __init__(
        self,
        s0: float,
        strike: float,
        expiry: float,
        r: float,
        mu: float,
        v0: float,
        kappa: float,
        theta: float,
        sigma: float,
        rho: float,
        n_steps: int,
        ticksize: float = 0.01,
    ):
        super().__init__(
            s0=s0,
            strike=strike,
            expiry=expiry,
            r=r,
            mu=mu,
            v0=v0,
            kappa=kappa,
            theta=theta,
            sigma=sigma,
            rho=rho,
            n_steps=n_steps,
            ticksize=ticksize,
        )
        self.action_space = Box(low=-1, high=0.0, shape=(1,))
        self.observation_space = Box(
            low=np.array([-np.inf, 0.0, 0.0, 0.0, 0.0, -1.0], dtype=np.float32),
            high=np.array([np.inf, 2.0, np.inf, 1.0, np.inf, 1.0], dtype=np.float32),
            shape=(6,),
        )


class HestonEnvDis(HestonEnvBase):

    def __init__(
        self,
        s0: float,
        strike: float,
        expiry: float,
        r: float,
        mu: float,
        v0: float,
        kappa: float,
        theta: float,
        sigma: float,
        rho: float,
        n_steps: int,
        ticksize: float = 0.01,
    ):
        super().__init__(
            s0=s0,
            strike=strike,
            expiry=expiry,
            r=r,
            mu=mu,
            v0=v0,
            kappa=kappa,
            theta=theta,
            sigma=sigma,
            rho=rho,
            n_steps=n_steps,
            ticksize=ticksize,
        )
        self.action_space = Discrete(100)
        self.observation_space = Box(
            low=np.array([-np.inf, 0.0, 0.0, 0.0, 0.0, -1.0], dtype=np.float32),
            high=np.array([np.inf, 2.0, np.inf, 1.0, np.inf, 1.0], dtype=np.float32),
            shape=(6,),
        )

    def step(self, action: Union[float, np.ndarray]):
        return super().step(-(action / 100))



from typing import Union

import numpy as np
from gymnasium import spaces
from gymnasium.spaces import Box, Discrete
from jax import vmap
from jaxfin.models.gbm import UnivGeometricBrownianMotion
from jaxfin.price_engine.black_scholes import delta_european, european_price

from ..base.call_env import HedgingEnvBase

v_delta_european = vmap(delta_european, in_axes=(0, None, None, None, None))


def flatten(fn):
    def wrapper(*args, **kwargs):
        return fn(*args, **kwargs).flatten()

    return wrapper


class BlackScholesEnvBase(HedgingEnvBase):
    metadata = {"render_modes": ["human"]}
    action_space: spaces.Space
    observation_space: spaces.Space

    def __init__(
        self,
        s0: float,
        strike: float,
        expiry: float,
        r: float,
        mu: float,
        sigma: float,
        n_steps: int,
        ticksize: float = 0.01,
    ):
        super().__init__(
            s0=s0,
            strike=strike,
            expiry=expiry,
            r=r,
            mu=mu,
            sigma=sigma,
            n_steps=n_steps,
            ticksize=ticksize,
        )

    def _generate_stock_path(self) -> np.ndarray:
        gbm = UnivGeometricBrownianMotion(self.s0, self.mu, self.sigma)

        return np.asarray(gbm.sample_paths(self.expiry, self.n_steps, 1))

    def _get_current_stock_vol(self, step: int) -> float:
        return self.sigma

    @flatten
    def _get_call_prices(self) -> np.ndarray:
        return np.asarray(
            [
                european_price(
                    self._stock_path[i],
                    self.strike,
                    self.expiry - i * self.dt,
                    self.sigma,
                    self.r,
                )
                for i in range(self.n_steps)
            ]
        )

    @flatten
    def _get_deltas(self) -> np.ndarray:
        return np.asarray(
            [
                v_delta_european(
                    self._stock_path[i],
                    self.strike,
                    self.expiry - i * self.dt,
                    self.sigma,
                    self.r,
                )
                for i in range(self.n_steps)
            ]
        )


class BlackScholesEnvCont(BlackScholesEnvBase):

    def __init__(
        self,
        s0: float,
        strike: float,
        expiry: float,
        r: float,
        mu: float,
        sigma: float,
        n_steps: int,
        ticksize: float = 0.01,
    ):
        super().__init__(
            s0=s0,
            strike=strike,
            expiry=expiry,
            r=r,
            mu=mu,
            sigma=sigma,
            n_steps=n_steps,
            ticksize=ticksize,
        )
        self.action_space = Box(low=-1.0, high=0.0, shape=(1,))
        self.observation_space = Box(
            low=np.array([-np.inf, 0.0, 0.0, 0.0, 0.0, -1.0], dtype=np.float32),
            high=np.array([np.inf, 2.0, np.inf, 1.0, np.inf, 1.0], dtype=np.float32),
            shape=(6,),
        )


class BlackScholesEnvDis(BlackScholesEnvBase):

    def __init__(
        self,
        s0: float,
        strike: float,
        expiry: float,
        r: float,
        mu: float,
        sigma: float,
        n_steps: int,
        ticksize: float = 0.01,
    ):
        super().__init__(
            s0=s0,
            strike=strike,
            expiry=expiry,
            r=r,
            mu=mu,
            sigma=sigma,
            n_steps=n_steps,
            ticksize=ticksize,
        )
        self.action_space = Discrete(100)
        self.observation_space = Box(
            low=np.array([-np.inf, 0.0, 0.0, 0.0, 0.0, -1.0], dtype=np.float32),
            high=np.array([np.inf, 2.0, np.inf, 1.0, np.inf, 1.0], dtype=np.float32),
            shape=(6,),
        )

    def step(self, action: Union[float, np.ndarray]):
        return super().step(-(action / 100))



from gymnasium.envs.registration import register

register(
    id="CallHedgingBSCont-v0", entry_point="hedging_gym.envs.bs_env:BlackScholesEnvCont"
)

register(
    id="CallHedgingBSDiscrete-v0",
    entry_point="hedging_gym.envs.bs_env:BlackScholesEnvDis",
)

register(
    id="CallHedgingHestonCont-v0",
    entry_point="hedging_gym.envs.heston_env:HestonEnvCont",
)

register(
    id="CallHedgingHestonDiscrete-v0",
    entry_point="hedging_gym.envs.heston_env:HestonEnvDis",
)

register(
    id="MargrabeHedgingCont-v0",
    entry_point="hedging_gym.envs.margrabe_env:MargrabeEnvCont",
)



from typing import Tuple

import jax.numpy as jnp
import numpy as np
from gymnasium import spaces
from gymnasium.spaces import Box, Discrete
from jax import vmap
from jaxfin.models.gbm import MultiGeometricBrownianMotion

from ..base.spread_env import SpreadHedgingEnvBase
from ..math import margrabe, margrabe_deltas


def flatten(fn):
    def wrapper(*args, **kwargs):
        return fn(*args, **kwargs).flatten()

    return wrapper


class MargrabeEnvBase(SpreadHedgingEnvBase):
    metadata = {"render_modes": ["human"]}
    action_space: spaces.Space
    observation_space: spaces.Space

    def __init__(
        self,
        s1_0: float,
        s2_0: float,
        expiry: float,
        r: float,
        mu_1: float,
        mu_2: float,
        sigma_1: float,
        sigma_2: float,
        corr: float,
        n_steps: int,
        tick_size: float = 0.01,
    ):
        super().__init__(
            s1_0=s1_0,
            s2_0=s2_0,
            strike=0.0,
            expiry=expiry,
            r=r,
            mu_1=mu_1,
            mu_2=mu_2,
            sigma_1=sigma_1,
            sigma_2=sigma_2,
            corr=corr,
            n_steps=n_steps,
            tick_size=tick_size,
        )

    def _generate_stock_path(self) -> Tuple[np.ndarray, np.ndarray]:
        s0 = jnp.array([self.s1_0, self.s2_0])
        means = jnp.array([self.mu_1, self.mu_2])
        vols = jnp.array([self.sigma_1, self.sigma_2])
        corr_matrix = jnp.array([[1, self.corr], [self.corr, 1]])
        gbm = MultiGeometricBrownianMotion(s0, means, vols, corr_matrix)

        stock_paths = gbm.sample_paths(self.expiry, self.n_steps, 1)

        return np.asarray(stock_paths[:, :, 0]), np.asarray(stock_paths[:, :, 1])

    def _get_current_stock_vols(self, step: int) -> Tuple[float, float]:
        return self.sigma_1, self.sigma_2

    @flatten
    def _get_spread_prices(self) -> np.ndarray:
        return np.asarray(
            [
                margrabe(
                    self._stock_path_2[i],
                    self._stock_path_1[i],
                    self.expiry - i * self.dt,
                    self.sigma_1,
                    self.sigma_2,
                    self.corr,
                )
                for i in range(self.n_steps)
            ]
        )

    def _get_deltas(self) -> Tuple[np.ndarray, np.ndarray]:
        deltas = np.asarray(
            [
                margrabe_deltas(
                    self._stock_path_2[i, 0],
                    self._stock_path_1[i, 0],
                    self.expiry - i * self.dt,
                    self.sigma_1,
                    self.sigma_2,
                    self.corr,
                )
                for i in range(self.n_steps)
            ]
        )
        print(deltas.shape)
        return deltas[:, 0], deltas[:, 1]


class MargrabeEnvCont(MargrabeEnvBase):

    def __init__(
        self,
        s1_0: float,
        s2_0: float,
        expiry: float,
        r: float,
        mu_1: float,
        mu_2: float,
        sigma_1: float,
        sigma_2: float,
        corr: float,
        n_steps: int,
        tick_size: float = 0.01,
    ):
        super().__init__(
            s1_0=s1_0,
            s2_0=s2_0,
            expiry=expiry,
            r=r,
            mu_1=mu_1,
            mu_2=mu_2,
            sigma_1=sigma_1,
            sigma_2=sigma_2,
            corr=corr,
            n_steps=n_steps,
            tick_size=tick_size,
        )
        self.action_space = Box(low=-1.0, high=1.0, shape=(2,))
        self.observation_space = Box(
            low=np.array(
                [-np.inf, 0.0, 0.0, 0.0, -1.0, -1.0, 0.0, -1.0, -1.0], dtype=np.float32
            ),
            high=np.array(
                [np.inf, 2.0, 2.0, np.inf, 1.0, 1.0, np.inf, 1.0, 1.0], dtype=np.float32
            ),
        )



from abc import ABCMeta, abstractmethod
from typing import Tuple

import numpy as np
from gymnasium import Env, spaces


class SpreadHedgingEnvBase(Env):
    metadata = {"render_modes": ["human"]}
    action_space: spaces.Space
    observation_space: spaces.Space

    __metaclass__ = ABCMeta

    def __init__(
        self,
        s1_0: float,
        s2_0: float,
        strike: float,
        expiry: float,
        r: float,
        mu_1: float,
        mu_2: float,
        sigma_1: float,
        sigma_2: float,
        corr: float,
        n_steps: int,
        tick_size: float = 0.01,
    ):
        self.s1_0 = s1_0
        self.s2_0 = s2_0
        self.strike = strike
        self.expiry = expiry
        self.r = r
        self.mu_1 = mu_1
        self.mu_2 = mu_2
        self.sigma_1 = sigma_1
        self.sigma_2 = sigma_2
        self.corr = corr
        self.n_steps = n_steps
        self.dt = expiry / n_steps
        self._current_step = 0

        self.render_mode = "human"

        self._stock_path_1 = np.array([])
        self._stock_path_2 = np.array([])
        self._spread_prices = np.array([])
        self._deltas_1 = np.array([])
        self._deltas_2 = np.array([])

        self._hedging_portfolio_value = 0.0
        self._current_hedging_delta_1 = 0.0
        self._current_hedging_delta_2 = 0.0
        self._previous_hedging_delta_1 = 0.0
        self._previous_hedging_delta_2 = 0.0
        self._back_account_value = 0.0
        self._ddelta_1 = 0.0
        self._ddelta_2 = 0.0

        self._epsilon = 1
        self._lambda = 0.1
        self._tick_size = tick_size

    def reset(self, seed=None, options=None):
        super().reset(seed=seed, options=options)

        self._current_step = 0
        self._stock_path_1, self._stock_path_2 = self._generate_stock_path()
        self._spread_prices = self._get_spread_prices()
        self._deltas_1, self._deltas_2 = self._get_deltas()
        self._hedging_portfolio_value = self._spread_prices[0]
        self._current_hedging_delta_1 = self._deltas_1[0]
        self._current_hedging_delta_2 = self._deltas_2[0]
        self._back_account_value = -(
            self._hedging_portfolio_value
            + self.current_hedging_delta_1 * self.s1_0
            + self.current_hedging_delta_2 * self.s2_0
            - self._get_transaction_costs(
                self.current_hedging_delta_1, self.current_hedging_delta_2
            )
        )

        observations = self._get_observations()
        infos = self._get_infos()

        return observations, infos

    def step(self, action: np.ndarray):
        self._current_step += 1
        delta_1 = action[0]
        delta_2 = action[1]

        done = self._current_step == (self.n_steps - 1)
        reward = self._calculate_reward()

        self._hedging_portfolio_value = self._calculate_hedging_portfolio_value(
            delta_1, delta_2
        )

        observations = self._get_observations()
        infos = self._get_infos()
        self._update_deltas(delta_1, delta_2)

        return observations, reward, done, False, infos

    def render(self, mode="human"):
        pass

    def _update_deltas(self, new_delta_1: float, new_delta_2: float):
        self._ddelta_1 = new_delta_1 - self.current_hedging_delta_1
        self._previous_hedging_delta_1 = self.current_hedging_delta_1
        self.current_hedging_delta_1 = new_delta_1
        self._ddelta_2 = new_delta_2 - self.current_hedging_delta_2
        self._previous_hedging_delta_2 = self.current_hedging_delta_2
        self.current_hedging_delta_2 = new_delta_2

    def _calculate_reward(self) -> float:
        if self._current_step == 1:
            return -self._get_transaction_costs(
                self.current_hedging_delta_1, self.current_hedging_delta_2
            )

        pnl = self._calculate_pnl()
        return pnl - self._lambda / 2 * pnl**2

    def _calculate_pnl(self) -> float:
        dv = (
            self._spread_prices[self._current_step]
            - self._spread_prices[self._current_step - 1]
        )
        ds_1 = (
            self._stock_path_1[self._current_step]
            - self._stock_path_1[self._current_step - 1]
        )[0]
        ds_2 = (
            self._stock_path_2[self._current_step]
            - self._stock_path_2[self._current_step - 1]
        )[0]
        ddelta_1 = self._ddelta_1
        ddelta_2 = self._ddelta_2

        if self._current_step == (self.n_steps - 1):
            liquidation_value = self._get_transaction_costs(
                self.current_hedging_delta_1, self.current_hedging_delta_2
            )
            return (
                dv
                + self.current_hedging_delta_1 * ds_1
                + self.current_hedging_delta_2 * ds_2
                - self._get_transaction_costs(ddelta_1, ddelta_2)
                - liquidation_value
            )

        return (
            dv
            + self.current_hedging_delta_1 * ds_1
            + self.current_hedging_delta_2 * ds_2
            - self._get_transaction_costs(ddelta_1, ddelta_2)
        )

    def _get_transaction_costs(self, ddelta_1: float, ddelta_2: float):
        return self._epsilon * self._tick_size * (
            np.abs(ddelta_1) + 0.01 * ddelta_1**2
        ) + self._epsilon * self._tick_size * (np.abs(ddelta_2) + 0.01 * ddelta_2**2)

    def _get_observations(self):
        log_price_strike = self._get_log_ratio()
        time_to_expiration = self.expiry
        vol_1, vol_2 = self._get_current_stock_vols(self._current_step)
        m_delta_1 = self._deltas_1[self._current_step]
        m_delta_2 = self._deltas_2[self._current_step]
        spread_price = self._spread_prices[self._current_step]
        return np.asarray(
            [
                log_price_strike,
                vol_1,
                vol_2,
                time_to_expiration,
                m_delta_1,
                m_delta_2,
                spread_price / self._spread_prices[0],
                self.current_hedging_delta_1,
                self.current_hedging_delta_2,
            ],
            dtype=np.float32,
        )

    def _get_infos(self):
        s1_t, s2_t = self._get_current_stock_prices()
        return {
            "price": self._spread_prices[self._current_step],
            "time_to_expiration": self.expiry - self._current_step * self.dt,
            "m_delta_1": self._deltas_1[self._current_step],
            "m_delta_2": self._deltas_2[self._current_step],
            "stock_1": s1_t,
            "stock_2": s2_t,
            "current_delta_1": self.current_hedging_delta_1,
            "current_delta_2": self.current_hedging_delta_2,
            "log(S_2/S_1)": self._get_log_ratio(),
            "hedge_portfolio_value": self._hedging_portfolio_value,
            "bank_account": self._back_account_value,
        }

    @property
    def current_hedging_delta_1(self):
        return self._current_hedging_delta_1

    @current_hedging_delta_1.setter
    def current_hedging_delta_1(self, new_hedge: float):
        self._current_hedging_delta_1 = new_hedge

    @property
    def current_hedging_delta_2(self):
        return self._current_hedging_delta_2
    
    @property
    def stock_path_1(self):
        return self._stock_path_1
    
    @property
    def stock_path_2(self):
        return self._stock_path_2

    @current_hedging_delta_2.setter
    def current_hedging_delta_2(self, new_hedge: float):
        self._current_hedging_delta_2 = new_hedge

    def _get_log_ratio(self):
        if self._current_step == -1:
            return np.log(self.s2_0 / self.s1_0)
        return np.log(
            self._stock_path_1[self._current_step, 0]
            / self._stock_path_2[self._current_step, 0]
        )

    @abstractmethod
    def _get_current_stock_vols(self, step: int) -> Tuple[float, float]:
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    def _generate_stock_path(self) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    def _get_spread_prices(self) -> np.ndarray:
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    def _get_deltas(self) -> Tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError("Subclasses must implement this method")

    def _calculate_hedging_portfolio_value(
        self, new_delta_1: float, new_delta_2: float
    ):
        stock_1, stock_2 = self._get_current_stock_prices()

        new_hedging_port_value = (
            self._back_account_value
            + self._current_hedging_delta_1 * stock_1
            + self._current_hedging_delta_2 * stock_2
        )

        if self._current_step == (self.n_steps - 1):
            liquidation_value = self._get_transaction_costs(
                self.current_hedging_delta_1, self.current_hedging_delta_2
            )
            transcation_costs = (
                self._get_transaction_costs(
                    self.current_hedging_delta_1, self.current_hedging_delta_2
                )
                + liquidation_value
            )
        else:
            transcation_costs = self._get_transaction_costs(
                self._ddelta_1, self._ddelta_2
            )

        self._back_account_value = (
            new_hedging_port_value
            - new_delta_1 * stock_1
            - new_delta_2 * stock_2
            - transcation_costs
        )

        return -new_hedging_port_value

    def _get_current_stock_prices(self) -> Tuple[float, float]:
        return (
            self._stock_path_1[self._current_step, 0],
            self._stock_path_2[self._current_step, 0],
        )



from abc import ABCMeta, abstractmethod
from typing import Union

import numpy as np
from gymnasium import Env, spaces


class HedgingEnvBase(Env):
    """
    Base abstract class for Call Hedging environment
    """

    metadata = {"render_modes": ["human"]}
    action_space: spaces.Space
    observation_space: spaces.Space

    __metaclass__ = ABCMeta

    def __init__(
        self,
        s0: float,
        strike: float,
        expiry: float,
        r: float,
        mu: float,
        sigma: float,
        n_steps: int,
        ticksize: float = 0.01,
    ):
        self.s0 = s0
        self.strike = strike
        self.expiry = expiry
        self.r = r
        self.mu = mu
        self.sigma = sigma
        self.n_steps = n_steps
        self.dt = expiry / n_steps
        self._current_step = 0

        self.render_mode = "human"

        self._stock_path = np.array([])
        self._call_prices = np.array([])
        self._deltas = np.array([])

        self._current_pnl = 0.0
        self._hedging_portfolio_value = 0.0
        self._current_hedging_delta = 0.0
        self._previous_hedging_delta = 0.0
        self._back_account_value = 0.0
        self._ddelta = 0.0

        self._epsilon = 1
        self._lambda = 0.1
        self._tick_size = ticksize

    def reset(self, seed=None, options=None):
        super().reset(seed=seed, options=options)

        self._current_step = 0
        self._current_pnl = 0.0
        self._stock_path = self._generate_stock_path()
        self._call_prices = self._get_call_prices()
        self._deltas = self._get_deltas()
        self._hedging_portfolio_value = self._call_prices[0]
        self._current_hedging_delta = -self._deltas[0]
        self._back_account_value = -(
            self._hedging_portfolio_value
            + self.current_hedging_delta * self.s0
            - self._get_transaction_costs(self.current_hedging_delta)
        )

        observations = self._get_observations()
        infos = self._get_infos()

        return observations, infos

    def _get_action(self, action: Union[np.ndarray, float]) -> float:
        if isinstance(action, np.ndarray):
            return action[0]
        return action

    def step(self, action: Union[np.ndarray, float]):
        action_f = self._get_action(action)

        self._current_step += 1

        done = self._current_step == (self.n_steps - 1)
        reward = self._calculate_reward()

        self._hedging_portfolio_value = self._calculate_hedging_portfolio_value(
            action_f
        )

        observations = self._get_observations()
        infos = self._get_infos()
        self._update_delta(action_f)

        return observations, reward, done, False, infos

    def render(self, mode="human"):
        pass

    def _update_delta(self, new_delta: float):
        self._ddelta = new_delta - self.current_hedging_delta
        self._previous_hedging_delta = self.current_hedging_delta
        self.current_hedging_delta = new_delta

    def _calculate_reward(self) -> float:
        if self._current_step == 1:
            return -self._get_transaction_costs(self.current_hedging_delta)

        pnl = self._calculate_pnl()
        self._current_pnl = pnl
        return pnl - self._lambda / 2 * pnl**2

    def _calculate_pnl(self) -> float:
        dv = (
            self._call_prices[self._current_step]
            - self._call_prices[self._current_step - 1]
        )
        ds = (
            self._stock_path[self._current_step]
            - self._stock_path[self._current_step - 1]
        )[0]
        ddelta = self._ddelta

        if self._current_step == (self.n_steps - 1):
            liquidation_value = self._get_transaction_costs(self.current_hedging_delta)
            return (
                dv
                + self.current_hedging_delta * ds
                - self._get_transaction_costs(ddelta)
                - liquidation_value
            )

        return (
            dv + self.current_hedging_delta * ds - self._get_transaction_costs(ddelta)
        )

    def _get_transaction_costs(self, ddelta: float):
        return self._epsilon * self._tick_size * (np.abs(ddelta) + 0.01 * ddelta**2)

    def _get_observations(self):
        log_price_strike = self._get_log_ratio()
        time_to_expiration = self.expiry
        bs_delta = self._deltas[self._current_step]
        call_price = self._call_prices[self._current_step]
        return np.asarray(
            [
                log_price_strike,
                self._get_current_stock_vol(self._current_step),
                time_to_expiration,
                bs_delta,
                call_price / self._call_prices[0],
                self.current_hedging_delta,
            ],
            dtype=np.float32,
        )

    def _get_infos(self):
        return {
            "price": self._call_prices[self._current_step],
            "time_to_expiration": self.expiry - self._current_step * self.dt,
            "bs_delta": self._deltas[self._current_step],
            "stock_price": self._get_current_stock_price(),
            "current_delta": self.current_hedging_delta,
            "log(S/K)": self._get_log_ratio(),
            "hedge_portfolio_value": self._hedging_portfolio_value,
            "bank_account": self._back_account_value,
            "current_pnl": self._current_pnl,
        }

    @property
    def current_hedging_delta(self):
        return self._current_hedging_delta

    @current_hedging_delta.setter
    def current_hedging_delta(self, new_hedge: float):
        self._current_hedging_delta = new_hedge

    @property
    def stock_path(self):
        return self._stock_path

    def _get_log_ratio(self):
        if self._current_step == -1:
            return np.log(self.s0 / self.strike)
        return np.log(self._stock_path[self._current_step, 0] / self.strike)

    @abstractmethod
    def _get_current_stock_vol(self, step: int) -> float:
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    def _generate_stock_path(self) -> np.ndarray:
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    def _get_call_prices(self) -> np.ndarray:
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    def _get_deltas(self) -> np.ndarray:
        raise NotImplementedError("Subclasses must implement this method")

    def _calculate_hedging_portfolio_value(self, new_delta: float):
        new_hedging_port_value = (
            self._back_account_value
            + self._current_hedging_delta * self._get_current_stock_price()
        )

        if self._current_step == (self.n_steps - 1):
            liquidation_value = self._get_transaction_costs(self.current_hedging_delta)
            transcation_costs = (
                self._get_transaction_costs(self.current_hedging_delta)
                + liquidation_value
            )
        else:
            transcation_costs = self._get_transaction_costs(self._ddelta)

        self._back_account_value = (
            new_hedging_port_value
            - new_delta * self._get_current_stock_price()
            - transcation_costs
        )

        return -new_hedging_port_value

    def _get_current_stock_price(self):
        return self._stock_path[self._current_step, 0]



"""
Taken from 
stable_baselines3/common/env_checker.py
used to check if the environment is compatible with Stable Baselines
"""

import warnings
from typing import Any, Dict, Union

import gymnasium as gym
import numpy as np
from gymnasium import spaces


def _is_numpy_array_space(space: spaces.Space) -> bool:
    """
    Returns False if provided space is not representable as a single numpy array
    (e.g. Dict and Tuple spaces return False)
    """
    return not isinstance(space, (spaces.Dict, spaces.Tuple))


def _starts_at_zero(space: Union[spaces.Discrete, spaces.MultiDiscrete]) -> bool:
    """
    Return False if a (Multi)Discrete space has a non-zero start.
    """
    return np.allclose(space.start, np.zeros_like(space.start))


def _check_non_zero_start(
    space: spaces.Space, space_type: str = "observation", key: str = ""
) -> None:
    """
    :param space: Observation or action space
    :param space_type: information about whether it is an observation or action space
        (for the warning message)
    :param key: When the observation space comes from a Dict space, we pass the
        corresponding key to have more precise warning messages. Defaults to "".
    """
    if isinstance(
        space, (spaces.Discrete, spaces.MultiDiscrete)
    ) and not _starts_at_zero(space):
        maybe_key = f"(key='{key}')" if key else ""
        warnings.warn(
            f"{type(space).__name__} {space_type} space {maybe_key} with a non-zero start (start={space.start}) "
            "is not supported by Stable-Baselines3. "
            f"You can use a wrapper or update your {space_type} space."
        )


def _check_image_input(observation_space: spaces.Box, key: str = "") -> None:
    """
    Check that the input will be compatible with Stable-Baselines
    when the observation is apparently an image.

    :param observation_space: Observation space
    :param key: When the observation space comes from a Dict space, we pass the
        corresponding key to have more precise warning messages. Defaults to "".
    """
    if observation_space.dtype != np.uint8:
        warnings.warn(
            f"It seems that your observation {key} is an image but its `dtype` "
            f"is ({observation_space.dtype}) whereas it has to be `np.uint8`. "
            "If your observation is not an image, we recommend you to flatten the observation "
            "to have only a 1D vector"
        )

    if np.any(observation_space.low != 0) or np.any(observation_space.high != 255):
        warnings.warn(
            f"It seems that your observation space {key} is an image but the "
            "upper and lower bounds are not in [0, 255]. "
            "Because the CNN policy normalize automatically the observation "
            "you may encounter issue if the values are not in that range."
        )


def _check_unsupported_spaces(
    env: gym.Env, observation_space: spaces.Space, action_space: spaces.Space
) -> None:
    """Emit warnings when the observation space or action space used is not supported by Stable-Baselines."""

    if isinstance(observation_space, spaces.Dict):
        nested_dict = False
        for key, space in observation_space.spaces.items():
            if isinstance(space, spaces.Dict):
                nested_dict = True
            _check_non_zero_start(space, "observation", key)

        if nested_dict:
            warnings.warn(
                "Nested observation spaces are not supported by Stable Baselines3 "
                "(Dict spaces inside Dict space). "
                "You should flatten it to have only one level of keys."
                "For example, `dict(space1=dict(space2=Box(), space3=Box()), spaces4=Discrete())` "
                "is not supported but `dict(space2=Box(), spaces3=Box(), spaces4=Discrete())` is."
            )

    if isinstance(observation_space, spaces.Tuple):
        warnings.warn(
            "The observation space is a Tuple, "
            "this is currently not supported by Stable Baselines3. "
            "However, you can convert it to a Dict observation space "
            "(cf. https://gymnasium.farama.org/api/spaces/composite/#dict). "
            "which is supported by SB3."
        )

    _check_non_zero_start(observation_space, "observation")

    if isinstance(observation_space, spaces.Sequence):
        warnings.warn(
            "Sequence observation space is not supported by Stable-Baselines3. "
            "You can pad your observation to have a fixed size instead.\n"
            "Note: The checks for returned values are skipped."
        )

    _check_non_zero_start(action_space, "action")

    if not _is_numpy_array_space(action_space):
        warnings.warn(
            "The action space is not based off a numpy array. Typically this means it's either a Dict or Tuple space. "
            "This type of action space is currently not supported by Stable Baselines 3. You should try to flatten the "
            "action using a wrapper."
        )


def _is_goal_env(env: gym.Env) -> bool:
    """
    Check if the env uses the convention for goal-conditioned envs (previously, the gym.GoalEnv interface)
    """
    # We need to unwrap the env since gym.Wrapper has the compute_reward method
    return hasattr(env.unwrapped, "compute_reward")


def _check_goal_env_obs(
    obs: dict, observation_space: spaces.Dict, method_name: str
) -> None:
    """
    Check that an environment implementing the `compute_rewards()` method
    (previously known as GoalEnv in gym) contains at least three elements,
    namely `observation`, `achieved_goal`, and `desired_goal`.
    """
    assert len(observation_space.spaces) >= 3, (
        "A goal conditioned env must contain at least 3 observation keys: `observation`, `achieved_goal`, and `desired_goal`. "
        f"The current observation contains {len(observation_space.spaces)} keys: {list(observation_space.spaces.keys())}"
    )

    for key in ["achieved_goal", "desired_goal"]:
        if key not in observation_space.spaces:
            raise AssertionError(
                f"The observation returned by the `{method_name}()` method of a goal-conditioned env requires the '{key}' "
                "key to be part of the observation dictionary. "
                f"Current keys are {list(observation_space.spaces.keys())}"
            )


def _check_goal_env_compute_reward(
    obs: Dict[str, Union[np.ndarray, int]],
    env: gym.Env,
    reward: float,
    info: Dict[str, Any],
) -> None:
    """
    Check that reward is computed with `compute_reward`
    and that the implementation is vectorized.
    """
    achieved_goal, desired_goal = obs["achieved_goal"], obs["desired_goal"]
    assert reward == env.compute_reward(  # type: ignore[attr-defined]
        achieved_goal, desired_goal, info
    ), "The reward was not computed with `compute_reward()`"

    achieved_goal, desired_goal = np.array(achieved_goal), np.array(desired_goal)
    batch_achieved_goals = np.array([achieved_goal, achieved_goal])
    batch_desired_goals = np.array([desired_goal, desired_goal])
    if isinstance(achieved_goal, int) or len(achieved_goal.shape) == 0:
        batch_achieved_goals = batch_achieved_goals.reshape(2, 1)
        batch_desired_goals = batch_desired_goals.reshape(2, 1)
    batch_infos = np.array([info, info])
    rewards = env.compute_reward(batch_achieved_goals, batch_desired_goals, batch_infos)  # type: ignore[attr-defined]
    assert rewards.shape == (
        2,
    ), f"Unexpected shape for vectorized computation of reward: {rewards.shape} != (2,)"
    assert (
        rewards[0] == reward
    ), f"Vectorized computation of reward differs from single computation: {rewards[0]} != {reward}"


def _check_obs(
    obs: Union[tuple, dict, np.ndarray, int],
    observation_space: spaces.Space,
    method_name: str,
) -> None:
    """
    Check that the observation returned by the environment
    correspond to the declared one.
    """
    if not isinstance(observation_space, spaces.Tuple):
        assert not isinstance(
            obs, tuple
        ), f"The observation returned by the `{method_name}()` method should be a single value, not a tuple"

    # The check for a GoalEnv is done by the base class
    if isinstance(observation_space, spaces.Discrete):
        # Since https://github.com/Farama-Foundation/Gymnasium/pull/141,
        # `sample()` will return a np.int64 instead of an int
        assert np.issubdtype(
            type(obs), np.integer
        ), f"The observation returned by `{method_name}()` method must be an int"
    elif _is_numpy_array_space(observation_space):
        assert isinstance(
            obs, np.ndarray
        ), f"The observation returned by `{method_name}()` method must be a numpy array"

    # Additional checks for numpy arrays, so the error message is clearer (see GH#1399)
    if isinstance(obs, np.ndarray):
        # check obs dimensions, dtype and bounds
        assert observation_space.shape == obs.shape, (
            f"The observation returned by the `{method_name}()` method does not match the shape "
            f"of the given observation space {observation_space}. "
            f"Expected: {observation_space.shape}, actual shape: {obs.shape}"
        )
        assert np.can_cast(obs.dtype, observation_space.dtype), (
            f"The observation returned by the `{method_name}()` method does not match the data type (cannot cast) "
            f"of the given observation space {observation_space}. "
            f"Expected: {observation_space.dtype}, actual dtype: {obs.dtype}"
        )
        if isinstance(observation_space, spaces.Box):
            lower_bounds, upper_bounds = observation_space.low, observation_space.high
            # Expose all invalid indices at once
            invalid_indices = np.where(
                np.logical_or(obs < lower_bounds, obs > upper_bounds)
            )
            if (obs > upper_bounds).any() or (obs < lower_bounds).any():
                message = (
                    f"The observation returned by the `{method_name}()` method does not match the bounds "
                    f"of the given observation space {observation_space}. \n"
                )
                message += f"{len(invalid_indices[0])} invalid indices: \n"

                for index in zip(*invalid_indices):
                    index_str = ",".join(map(str, index))
                    message += (
                        f"Expected: {lower_bounds[index]} <= obs[{index_str}] <= {upper_bounds[index]}, "
                        f"actual value: {obs[index]} \n"
                    )

                raise AssertionError(message)

    assert observation_space.contains(obs), (
        f"The observation returned by the `{method_name}()` method "
        f"does not match the given observation space {observation_space}"
    )


def _check_box_obs(observation_space: spaces.Box, key: str = "") -> None:
    """
    Check that the observation space is correctly formatted
    when dealing with a ``Box()`` space. In particular, it checks:
    - that the dimensions are big enough when it is an image, and that the type matches
    - that the observation has an expected shape (warn the user if not)
    """
    # If image, check the low and high values, the type and the number of channels
    # and the shape (minimal value)
    if len(observation_space.shape) == 3:
        _check_image_input(observation_space, key)

    if len(observation_space.shape) not in [1, 3]:
        warnings.warn(
            f"Your observation {key} has an unconventional shape (neither an image, nor a 1D vector). "
            "We recommend you to flatten the observation "
            "to have only a 1D vector or use a custom policy to properly process the data."
        )


def _check_returned_values(
    env: gym.Env, observation_space: spaces.Space, action_space: spaces.Space
) -> None:
    """
    Check the returned values by the env when calling `.reset()` or `.step()` methods.
    """
    # because env inherits from gymnasium.Env, we assume that `reset()` and `step()` methods exists
    reset_returns = env.reset()
    assert isinstance(reset_returns, tuple), "`reset()` must return a tuple (obs, info)"
    assert (
        len(reset_returns) == 2
    ), f"`reset()` must return a tuple of size 2 (obs, info), not {len(reset_returns)}"
    obs, info = reset_returns
    assert isinstance(
        info, dict
    ), f"The second element of the tuple return by `reset()` must be a dictionary not {info}"

    if _is_goal_env(env):
        # Make mypy happy, already checked
        assert isinstance(observation_space, spaces.Dict)
        _check_goal_env_obs(obs, observation_space, "reset")
    elif isinstance(observation_space, spaces.Dict):
        assert isinstance(
            obs, dict
        ), "The observation returned by `reset()` must be a dictionary"

        if not obs.keys() == observation_space.spaces.keys():
            raise AssertionError(
                "The observation keys returned by `reset()` must match the observation "
                f"space keys: {obs.keys()} != {observation_space.spaces.keys()}"
            )

        for key in observation_space.spaces.keys():
            try:
                _check_obs(obs[key], observation_space.spaces[key], "reset")
            except AssertionError as e:
                raise AssertionError(
                    f"Error while checking key={key}: " + str(e)
                ) from e
    else:
        _check_obs(obs, observation_space, "reset")

    # Sample a random action
    action = action_space.sample()
    data = env.step(action)

    assert len(data) == 5, (
        "The `step()` method must return five values: "
        f"obs, reward, terminated, truncated, info. Actual: {len(data)} values returned."
    )

    # Unpack
    obs, reward, terminated, truncated, info = data

    if isinstance(observation_space, spaces.Dict):
        assert isinstance(
            obs, dict
        ), "The observation returned by `step()` must be a dictionary"

        # Additional checks for GoalEnvs
        if _is_goal_env(env):
            # Make mypy happy, already checked
            assert isinstance(observation_space, spaces.Dict)
            _check_goal_env_obs(obs, observation_space, "step")
            _check_goal_env_compute_reward(obs, env, float(reward), info)

        if not obs.keys() == observation_space.spaces.keys():
            raise AssertionError(
                "The observation keys returned by `step()` must match the observation "
                f"space keys: {obs.keys()} != {observation_space.spaces.keys()}"
            )

        for key in observation_space.spaces.keys():
            try:
                _check_obs(obs[key], observation_space.spaces[key], "step")
            except AssertionError as e:
                raise AssertionError(
                    f"Error while checking key={key}: " + str(e)
                ) from e

    else:
        _check_obs(obs, observation_space, "step")

    # We also allow int because the reward will be cast to float
    assert isinstance(
        reward, (float, int)
    ), "The reward returned by `step()` must be a float"
    assert isinstance(terminated, bool), "The `terminated` signal must be a boolean"
    assert isinstance(truncated, bool), "The `truncated` signal must be a boolean"
    assert isinstance(
        info, dict
    ), "The `info` returned by `step()` must be a python dictionary"

    # Goal conditioned env
    if _is_goal_env(env):
        # for mypy, env.unwrapped was checked by _is_goal_env()
        assert hasattr(env, "compute_reward")
        assert reward == env.compute_reward(
            obs["achieved_goal"], obs["desired_goal"], info
        )


def _check_spaces(env: gym.Env) -> None:
    """
    Check that the observation and action spaces are defined and inherit from spaces.Space. For
    envs that follow the goal-conditioned standard (previously, the gym.GoalEnv interface) we check
    the observation space is gymnasium.spaces.Dict
    """
    gym_spaces = "cf. https://gymnasium.farama.org/api/spaces/"

    assert hasattr(
        env, "observation_space"
    ), f"You must specify an observation space ({gym_spaces})"
    assert hasattr(
        env, "action_space"
    ), f"You must specify an action space ({gym_spaces})"

    assert isinstance(
        env.observation_space, spaces.Space
    ), f"The observation space must inherit from gymnasium.spaces ({gym_spaces})"
    assert isinstance(
        env.action_space, spaces.Space
    ), f"The action space must inherit from gymnasium.spaces ({gym_spaces})"

    if _is_goal_env(env):
        print(
            "We detected your env to be a GoalEnv because `env.compute_reward()` was defined.\n"
            "If it's not the case, please rename `env.compute_reward()` to something else to avoid False positives."
        )
        assert isinstance(env.observation_space, spaces.Dict), (
            "Goal conditioned envs (previously gym.GoalEnv) require the observation space to be gymnasium.spaces.Dict.\n"
            "Note: if your env is not a GoalEnv, please rename `env.compute_reward()` "
            "to something else to avoid False positive."
        )


# Check render cannot be covered by CI
def _check_render(env: gym.Env, warn: bool = False) -> None:  # pragma: no cover
    """
    Check the instantiated render mode (if any) by calling the `render()`/`close()`
    method of the environment.

    :param env: The environment to check
    :param warn: Whether to output additional warnings
    :param headless: Whether to disable render modes
        that require a graphical interface. False by default.
    """
    render_modes = env.metadata.get("render_modes")
    if render_modes is None:
        if warn:
            warnings.warn(
                "No render modes was declared in the environment "
                "(env.metadata['render_modes'] is None or not defined), "
                "you may have trouble when calling `.render()`"
            )

    # Only check currrent render mode
    if env.render_mode:
        env.render()
    env.close()


def check_env(env: gym.Env, warn: bool = True, skip_render_check: bool = True) -> None:
    """
    Check that an environment follows Gym API.
    This is particularly useful when using a custom environment.
    Please take a look at https://gymnasium.farama.org/api/env/
    for more information about the API.

    It also optionally check that the environment is compatible with Stable-Baselines.

    :param env: The Gym environment that will be checked
    :param warn: Whether to output additional warnings
        mainly related to the interaction with Stable Baselines
    :param skip_render_check: Whether to skip the checks for the render method.
        True by default (useful for the CI)
    """
    assert isinstance(
        env, gym.Env
    ), "Your environment must inherit from the gymnasium.Env class cf. https://gymnasium.farama.org/api/env/"

    # ============= Check the spaces (observation and action) ================
    _check_spaces(env)

    # Define aliases for convenience
    observation_space = env.observation_space
    action_space = env.action_space

    try:
        env.reset(seed=0)
    except TypeError as e:
        raise TypeError("The reset() method must accept a `seed` parameter") from e

    # Warn the user if needed.
    # A warning means that the environment may run but not work properly with Stable Baselines algorithms
    if warn:
        _check_unsupported_spaces(env, observation_space, action_space)

        obs_spaces = (
            observation_space.spaces
            if isinstance(observation_space, spaces.Dict)
            else {"": observation_space}
        )
        for key, space in obs_spaces.items():
            if isinstance(space, spaces.Box):
                _check_box_obs(space, key)

        # Check for the action space, it may lead to hard-to-debug issues
        if isinstance(action_space, spaces.Box) and (
            np.any(np.abs(action_space.low) != np.abs(action_space.high))
            or np.any(action_space.low != -1)
            or np.any(action_space.high != 1)
        ):
            warnings.warn(
                "We recommend you to use a symmetric and normalized Box action space (range=[-1, 1]) "
                "cf. https://stable-baselines3.readthedocs.io/en/master/guide/rl_tips.html"
            )

        if isinstance(action_space, spaces.Box):
            assert np.all(
                np.isfinite(np.array([action_space.low, action_space.high]))
            ), "Continuous action space must have a finite lower and upper bound"

        if isinstance(action_space, spaces.Box) and action_space.dtype != np.dtype(
            np.float32
        ):
            warnings.warn(
                f"Your action space has dtype {action_space.dtype}, we recommend using np.float32 to avoid cast errors."
            )

    # If Sequence observation space, do not check the observation any further
    if isinstance(observation_space, spaces.Sequence):
        return

    # ============ Check the returned values ===============
    _check_returned_values(env, observation_space, action_space)

    # ==== Check the render method and the declared render modes ====
    if not skip_render_check:
        _check_render(env, warn)  # pragma: no cover
